import { includeBytes } from "fastly:experimental";
import { ObjectStore } from "fastly:object-store";
import { AssetManager } from "./asset-manager.js";
import { IncludeBytesStoreEntry } from "../server/object-store/include-bytes-store-entry.js";
const decoder = new TextDecoder();
export class ContentInlineAsset {
    constructor(metadata, bytes) {
        this.isInline = true;
        this.metadata = metadata;
        this.bytes = bytes;
    }
    get assetKey() {
        return this.metadata.assetKey;
    }
    getStoreEntry() {
        return Promise.resolve(new IncludeBytesStoreEntry(this.bytes));
    }
    getBytes() {
        return this.bytes;
    }
    getText() {
        if (!this.metadata.text) {
            throw new Error("Can't getText() for non-text content");
        }
        return decoder.decode(this.bytes);
    }
    getMetadata() {
        return this.metadata;
    }
}
export class ContentObjectStoreAsset {
    constructor(metadata, objectStoreName) {
        this.isInline = false;
        this.metadata = metadata;
        this.objectStoreName = objectStoreName;
    }
    get assetKey() {
        return this.metadata.assetKey;
    }
    async getStoreEntry() {
        const objectStore = new ObjectStore(this.objectStoreName);
        let retries = 3;
        while (retries > 0) {
            const entry = await objectStore.get(this.metadata.objectStoreKey);
            if (entry != null) {
                return entry;
            }
            // Note the null does NOT mean 404. The fact that we are here means
            // metadata exists for this path, and we're just trying to get the data from
            // the object store.
            // So if we're here then the data is either not available yet (in which case
            // we can wait just a bit and try again), or the data was deleted.
            retries--;
            // We're going to wait 250ms/500ms/750ms and try again.
            const delay = (3 - retries) * 250;
            await new Promise(resolve => setTimeout(resolve, delay));
        }
        throw new Error("Asset could not be retrieved from the Object Store.");
    }
    getBytes() {
        throw new Error("Can't getBytes() for Object Store asset");
    }
    getText() {
        throw new Error("Can't getText() for Object Store asset");
    }
    getMetadata() {
        return this.metadata;
    }
}
export class ContentAssets extends AssetManager {
    constructor(objectStoreName, contentAssetMetadataMap) {
        super();
        for (const [assetKey, metadata] of Object.entries(contentAssetMetadataMap)) {
            let asset;
            if (metadata.isInline) {
                const bytes = includeBytes(metadata.staticFilePath);
                asset = new ContentInlineAsset(metadata, bytes);
            }
            else {
                if (objectStoreName == null) {
                    throw new Error("Unexpected! Object Store should be specified!!");
                }
                asset = new ContentObjectStoreAsset(metadata, objectStoreName);
            }
            this.setAsset(assetKey, asset);
        }
    }
}
